import tkinter as tk
from typing import Callable

import customtkinter as ctk

import flowmatic
from flowmatic import gui
from flowmatic.flows.flow import Flow
from flowmatic.gui.screens.screen import FlowScreen, Screen
from ...forms import FormValidation
from ...forms.fields.field import Field


class FormScreen(Screen):
    field_list: list[Field] | None
    fields: dict[str, Field]
    validations: list[FormValidation] | None = None
    __message: tk.StringVar

    @property
    def message(self) -> str:
        return self.__message.get()

    @message.setter
    def message(self, value: str) -> None:
        """Set message.

        Args:
            value (str): Message."""
        self.__message.set(value)

    def __init__(
        self,
    ) -> None:
        self.fields = {}
        self.__message = tk.StringVar(value="")

    @property
    def values(self) -> dict[str, str | int | list[str]]:
        """Get values from fields.

        Returns:
            dict[str, str]: Values from fields."""
        return {
            label: field.value for label, field in self.fields.items() if field.value
        }

    def validate(self) -> None:
        if not self.validations:
            return

        for validation in self.validations:
            validation.validate(self.values)

    def build(  # pylint: disable=arguments-differ
        self,
        *,
        title: str,
        validations: list[FormValidation] | None = None,
        on_submit: Callable[[], None],
        submit_text: str = "Submit",
        back_command: Callable[[], None] | None = None,
        fields: list[Field] | None = None,
        field_factory: Callable[..., list[Field]] | None = None,
    ) -> None:
        """Build screen.

        Args:

            title (str): Title of screen.
            validate_command (Callable[[], None]): Command to run when validate button is pressed.
            validate_text (str, optional): Text of validate button. Defaults to "Submit".
            back_command (Callable[[], None], optional): Command to run when back button is pressed.
                 Defaults to None.
        """
        # Variables
        self.field_list = (fields or []) + (field_factory() if field_factory else [])
        self.validations = validations

        def validate_command() -> None:
            try:
                self.validate()
            except AssertionError as error:
                self.message = str(error)
                flowmatic.update_gui()
            else:
                on_submit()

        # Title
        ctk.CTkLabel(self.master, text=title).pack(**gui.pack_defaults)
        # Fields
        if self.field_list:
            fields_frame = ctk.CTkFrame(self.master, width=gui.WIDTH)
            fields_frame.pack(**gui.pack_defaults)
            for field in self.field_list:
                self.fields[field.label] = field(fields_frame).build(
                    **gui.pack_defaults
                )
        # Message
        ctk.CTkLabel(self.master, textvariable=self.__message).pack(**gui.pack_defaults)
        # Buttons
        ctk.CTkButton(self.master, text=submit_text, command=validate_command).pack(
            **gui.pack_defaults, side=tk.RIGHT
        )
        if back_command:
            ctk.CTkButton(self.master, text="Back", command=back_command).pack(
                **gui.pack_defaults, side=tk.LEFT
            )


class FlowFormScreen(FormScreen, FlowScreen):
    def __init__(self, flow: Flow) -> None:
        self.flow = flow
        super().__init__()



from typing import Protocol, Self
import tkinter as tk

from ...flows.flow import Flow


class Screen(Protocol):
    """Screen protocol.

    Args:
        master (tk.Tk): Master window.

    Methods:
        build: Build screen."""

    master: tk.Tk

    def __init__(self) -> None:
        ...

    def __call__(self, master: tk.Tk) -> Self:
        self.master = master  # pylint: disable=attribute-defined-outside-init
        return self

    def build(self) -> None:  # pylint: disable=missing-function-docstring
        ...


class FlowScreen(Screen):
    flow: Flow

    def __init__(self, flow: Flow) -> None:
        self.flow = flow
        super().__init__()

    def next(self) -> None:
        self.flow.next()

    def previous(self) -> None:
        self.flow.previous()

    def build(self) -> None:
        ...



from .gui import GUI, TKGUI
from .gui.screens.screen import Screen
from .util.settings import Settings
from .util.user import User
from .flows.flow import Flow


class App:
    """Singleton for accessing GUI, settings, user and other app-wide objects.

    Attributes:
        gui (GUI): GUI interface.
        settings (Settings): Settings.
        user (User): User.

    Args:
        title (str, optional): Title of the app. Defaults to "Appy".
    """

    title: str = "FlowMatic"
    gui: GUI = None  # type: ignore
    settings: Settings
    user: User

    start_screen: Screen

    def __init__(self, start_screen: type[Screen]):
        self.gui = self.gui or TKGUI(
            title=self.title, start_screen=start_screen, geometry="1280x720"
        )
        self.start_screen = start_screen()
        self.settings = Settings.load("files/settings.json")
        self.user = User("")

    def start(
        self,
    ):
        """Start app.
        Args:
            screen (type[Screen]): Screen to start with."""
        self.gui.start()

    def show_screen(self, screen: Screen):
        """Show start screen.
        Args:
            screen (type[Screen]): Screen to show."""
        self.gui.switch_screen(screen)

    def start_flow(self, flow: Flow):
        """Start flow.
        Args:
            flow (type[Flow]): Flow to start."""
        self.gui.start_flow(flow)

    def quit(self):
        self.gui.quit()



"""GUI Interface for Appy.
    Provides Interface to GUI elements and screens. 
"""


from . import style
from .gui import GUI
from .tkgui import TKGUI

pack_defaults = {**style.paddings, "expand": True}
WIDTH = 1280
HEIGHT = 720
